package org.jboss.cache.factories.annotations;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

/**
 * Used with the {@link org.jboss.cache.factories.annotations.Inject} annotation to specify the name of a component to inject.
 * Note that you still need to annotate your method with {@link org.jboss.cache.factories.annotations.Inject} for this
 * annotation to be of any use.  You can also combine injections of named and singleton components, see example below.
 * <p/>
 * Usage example:
 * <p/>
 * <pre>
 *       public class MyClass
 *       {
 *          private Map blahMap;
 *          private Map anotherMap;
 *          private List someList;
 *          private TransactionManager tm;
 * <p/>
 *          &amp; Inject
 *          public void setMyMap(&amp;ComponentName("BlahMap") Map blahMap)
 *          {
 *             this.blahMap = blahMap;
 *          }
 *          &amp; Inject
 *          public void injectMoreStuff(&amp;ComponentName("AnotherMap") Map anotherMap, &amp;ComponentName("SomeList") List someList, TransactionManager singletonTM)
 *          {
 *             this.anotherMap = anotherMap;
 *             this.someList = someList;
 *             this.tm = singletonTM;
 *          }
 *       }
 * <p/>
 * </pre>
 *
 * @author Manik Surtani (<a href="mailto:manik@jboss.org">manik@jboss.org</a>)
 * @see org.jboss.cache.factories.annotations.Inject
 * @since 2.1.0
 */
// ensure this annotation is available at runtime.
@Retention(RetentionPolicy.RUNTIME)

// only applies to fields.
@Target(ElementType.PARAMETER)

public @interface ComponentName
{
   String value();
}
