package org.jboss.cache.interceptors;

import org.jboss.cache.InvocationContext;
import org.jboss.cache.factories.annotations.ComponentName;
import org.jboss.cache.factories.annotations.Inject;
import org.jboss.cache.lock.NodeLock;

import javax.transaction.Transaction;
import java.util.List;
import java.util.Map;

/**
 * When a call returns, unlocks all locks held by the current thread in the
 * LockTable. This is a no-op if a transaction is used.
 *
 * @author Bela Ban
 * @version $Id: UnlockInterceptor.java 4968 2008-01-03 19:17:09Z manik.surtani@jboss.com $
 */
public class UnlockInterceptor extends Interceptor
{

   Map<Thread, List<NodeLock>> lockTable = null;
//   ThreadLocal<List<NodeLock>> lockTable;

   public UnlockInterceptor()
   {
      initLogger();
   }

   @Inject
   private void injectDependencies(@ComponentName("LockTable")Map<Thread, List<NodeLock>> lockTable)
//         private void injectDependencies(@ComponentName("LockTable")ThreadLocal<List<NodeLock>> lockTable)
   {
      this.lockTable = lockTable;
   }

   public Object invoke(InvocationContext ctx) throws Throwable
   {
      try
      {
         return nextInterceptor(ctx);
      }
      catch (Throwable th)
      {
         throw th;
      }
      finally
      {
         if (ctx.getOptionOverrides() == null || !ctx.getOptionOverrides().isSuppressLocking())
         {
            Transaction tx = ctx.getTransaction();
            if (tx != null && isValid(tx))
            {
               // if (trace) log.trace("Do not do anything; we have a transaction running or node locking is optimistic.");
            }
            else
            { // no TX
               Thread currentThread = Thread.currentThread();
               List<NodeLock> locks = lockTable.get(currentThread);
               if (trace) log.trace("Attempting to release locks on current thread.  Lock table is " + lockTable);

               if (locks != null && locks.size() > 0)
               {
                  releaseLocks(locks, currentThread);
                  lockTable.remove(currentThread);
               }
            }
         }
      }
   }

   private void releaseLocks(List<NodeLock> locks, Thread currentThread)
   {
      NodeLock[] locksArray = locks.toArray(new NodeLock[]{});
      for (int i = locksArray.length - 1; i > -1; i--)
      {
         if (trace) log.trace("releasing lock for " + locksArray[i].getFqn() + ": " + locksArray[i]);
         locksArray[i].release(currentThread);
      }
   }


}
