package org.jboss.cache.eviction;

import static org.testng.AssertJUnit.assertEquals;
import static org.testng.AssertJUnit.assertTrue;
import static org.testng.AssertJUnit.fail;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.Region;
import org.jboss.cache.RegionManager;
import org.jboss.cache.misc.TestingUtil;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

/**
 * Unit tests for LRUAlgorithm.
 *
 * @author Ben Wang, Feb 11, 2004
 * @author Daniel Huang (dhuang@jboss.org)
 */
@Test(groups = {"functional"})
public class LRUAlgorithmTest
{
   RegionManager regionManager_;
   LRUAlgorithm algo_;
   LRUConfiguration config_;
   Log log = LogFactory.getLog(LRUAlgorithm.class);

   @BeforeMethod(alwaysRun = true)
   public void setUp() throws Exception
   {
      algo_ = new LRUAlgorithm();
      config_ = new LRUConfiguration();
      config_.setEvictionPolicyClass(DummyEvictionPolicy.class.getName());
      // We have to setCache timeToLiveSeconds!!
      config_.setTimeToLiveSeconds(0);

      regionManager_ = new RegionManager();
      regionManager_.getRegion("/a/b", true).setEvictionPolicy(config_);
      /*
      try {
         Thread.sleep(10000);
      } catch (InterruptedException e) {
         e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
      }
      */
   }

   /**
    * maxNodes = 1. Eception is evictFromCacheNode. Should be commented for now.
    */
   public void XtestEvictException()
   {
      Fqn fqn1 = Fqn.fromString("/a/b/c");
      Fqn fqn2 = Fqn.fromString("/a/b/d");
      Fqn fqn3 = Fqn.fromString("/a/b/e");
      Region region = regionManager_.getRegion("/a/b", true);
      LRUConfiguration config = (LRUConfiguration) region.getEvictionPolicyConfig();
      config.setMaxNodes(1);
      region.putNodeEvent(new EvictedEventNode(fqn1, NodeEventType.ADD_NODE_EVENT));
      region.putNodeEvent(new EvictedEventNode(fqn2, NodeEventType.ADD_NODE_EVENT));

      try
      {
         algo_.process(region);
      }
      catch (EvictionException e)
      {
         fail("testMaxNode: process failed " + e);
         e.printStackTrace();
      }
      assertEquals("Queue size should be ", 1, algo_.getEvictionQueue().getNumberOfNodes());

      region.putNodeEvent(new EvictedEventNode(fqn2, NodeEventType.ADD_NODE_EVENT));
      region.putNodeEvent(new EvictedEventNode(fqn3, NodeEventType.ADD_NODE_EVENT));

      try
      {
         algo_.process(region);
      }
      catch (EvictionException e)
      {
         fail("testMaxNode: process failed " + e);
         e.printStackTrace();
      }
      assertEquals("Queue size should be ", 1, algo_.getEvictionQueue().getNumberOfNodes());
   }


   /**
    * maxNodes = 0 case
    */
   public void testMaxNode1()
   {
      Fqn fqn1 = Fqn.fromString("/a/b/c");
      Fqn fqn2 = Fqn.fromString("/a/b/d");
      Region region = regionManager_.getRegion("/a/b", true);
      LRUConfiguration config = (LRUConfiguration) region.getEvictionPolicyConfig();
      config.setMaxNodes(0);
      region.putNodeEvent(new EvictedEventNode(fqn1, NodeEventType.ADD_NODE_EVENT));
      region.putNodeEvent(new EvictedEventNode(fqn2, NodeEventType.ADD_NODE_EVENT));

      try
      {
         algo_.process(region);
      }
      catch (EvictionException e)
      {
         fail("testMaxNode: process failed " + e);
         e.printStackTrace();
      }
      assertEquals("Queue size should be ", 2, algo_.getEvictionQueue().getNumberOfNodes());

   }

   /**
    * maxNodes = 1
    */
   public void testMaxNode2()
   {
      Fqn fqn1 = Fqn.fromString("/a/b/c");
      Fqn fqn2 = Fqn.fromString("/a/b/d");
      Fqn fqn3 = Fqn.fromString("/a/b/e");
      Region region = regionManager_.getRegion("/a/b", true);
      LRUConfiguration config = (LRUConfiguration) region.getEvictionPolicyConfig();
      config.setMaxNodes(1);
      region.putNodeEvent(new EvictedEventNode(fqn1, NodeEventType.ADD_NODE_EVENT));
      region.putNodeEvent(new EvictedEventNode(fqn2, NodeEventType.ADD_NODE_EVENT));

      try
      {
         algo_.process(region);
      }
      catch (EvictionException e)
      {
         fail("testMaxNode: process failed " + e);
         e.printStackTrace();
      }
      assertEquals("Queue size should be ", 1, algo_.getEvictionQueue().getNumberOfNodes());

      region.putNodeEvent(new EvictedEventNode(fqn2, NodeEventType.ADD_NODE_EVENT));
      region.putNodeEvent(new EvictedEventNode(fqn3, NodeEventType.ADD_NODE_EVENT));

      try
      {
         algo_.process(region);
      }
      catch (EvictionException e)
      {
         fail("testMaxNode: process failed " + e);
         e.printStackTrace();
      }
      assertEquals("Queue size should be ", 1, algo_.getEvictionQueue().getNumberOfNodes());
   }

   /**
    * TimeToIdleSeconds = 0
    */
   public void testIdleTimeSeconds1()
   {
      Fqn fqn1 = Fqn.fromString("/a/b/c");
      Fqn fqn2 = Fqn.fromString("/a/b/d");
      Region region = regionManager_.getRegion("/a/b", true);
      LRUConfiguration config = (LRUConfiguration) region.getEvictionPolicyConfig();
      config.setMaxNodes(0);
      config.setTimeToLiveSeconds(0);
      region.putNodeEvent(new EvictedEventNode(fqn1, NodeEventType.ADD_NODE_EVENT));
      region.putNodeEvent(new EvictedEventNode(fqn2, NodeEventType.ADD_NODE_EVENT));
      TestingUtil.sleepThread(2000);
      try
      {
         algo_.process(region);
      }
      catch (EvictionException e)
      {
         fail("testMaxNode: process failed " + e);
         e.printStackTrace();
      }
      assertEquals("Queue size should be ", 2, algo_.getEvictionQueue().getNumberOfNodes());

   }

   /**
    * TimeToIdleSeconds = 1
    */
   public void testIdleTimeSeconds2()
   {
      Fqn fqn1 = Fqn.fromString("/a/b/c");
      Fqn fqn2 = Fqn.fromString("/a/b/d");
      Fqn fqn3 = Fqn.fromString("/a/b/e");
      Region region = regionManager_.getRegion("/a/b", true);
      LRUConfiguration config = (LRUConfiguration) region.getEvictionPolicyConfig();
      config.setMaxNodes(0);
      config.setTimeToLiveSeconds(1);
      region.putNodeEvent(new EvictedEventNode(fqn1, NodeEventType.ADD_NODE_EVENT));
      region.putNodeEvent(new EvictedEventNode(fqn2, NodeEventType.ADD_NODE_EVENT));
      region.putNodeEvent(new EvictedEventNode(fqn3, NodeEventType.ADD_NODE_EVENT));

      try
      {
         algo_.process(region);
      }
      catch (EvictionException e)
      {
         fail("testMaxNode: process failed " + e);
         e.printStackTrace();
      }
      assertEquals("Queue size #1: ", 3, algo_.getEvictionQueue().getNumberOfNodes());
      TestingUtil.sleepThread(2000);
      try
      {
         algo_.process(region);
      }
      catch (EvictionException e)
      {
         fail("testMaxNode: process failed " + e);
         e.printStackTrace();
      }
      assertEquals("Queue size #2: ", 0, algo_.getEvictionQueue().getNumberOfNodes());
   }

   /**
    * TimeToIdleSeconds = 1 with node visited in between.
    */
   public void testIdleTimeSeconds3()
   {
      Fqn fqn1 = Fqn.fromString("/a/b/c");
      Fqn fqn2 = Fqn.fromString("/a/b/d");
      Fqn fqn3 = Fqn.fromString("/a/b/e");
      Region region = regionManager_.getRegion("/a/b", true);
      LRUConfiguration config = (LRUConfiguration) region.getEvictionPolicyConfig();
      config.setMaxNodes(0);
      config.setTimeToLiveSeconds(1);
      region.putNodeEvent(new EvictedEventNode(fqn1, NodeEventType.ADD_NODE_EVENT));
      region.putNodeEvent(new EvictedEventNode(fqn2, NodeEventType.ADD_NODE_EVENT));
      region.putNodeEvent(new EvictedEventNode(fqn3, NodeEventType.ADD_NODE_EVENT));

      try
      {
         algo_.process(region);
      }
      catch (EvictionException e)
      {
         fail("testMaxNode: process failed " + e);
         e.printStackTrace();
      }
      assertEquals("Queue size #1: ", 3, algo_.getEvictionQueue().getNumberOfNodes());
      TestingUtil.sleepThread(2000);
      region.putNodeEvent(new EvictedEventNode(fqn2, NodeEventType.VISIT_NODE_EVENT));
      try
      {
         algo_.process(region);
      }
      catch (EvictionException e)
      {
         fail("testMaxNode: process failed " + e);
         e.printStackTrace();
      }
      assertEquals("Queue size #2: ", 1, algo_.getEvictionQueue().getNumberOfNodes());
   }

   /**
    * MaxAgeSeconds = 1 with 3 nodes.
    *
    * @throws Exception
    */
   public void testMaxAgeSeconds1() throws Exception
   {
      Fqn fqn1 = Fqn.fromString("/a/b/c");
      Fqn fqn2 = Fqn.fromString("/a/b/d");
      Fqn fqn3 = Fqn.fromString("/a/b/e");
      Region region = regionManager_.getRegion("/a/b", true);
      LRUConfiguration config = (LRUConfiguration) region.getEvictionPolicyConfig();
      config.setMaxNodes(0);
      config.setTimeToLiveSeconds(0);
      config.setMaxAgeSeconds(1);
      region.putNodeEvent(new EvictedEventNode(fqn1, NodeEventType.ADD_NODE_EVENT));
      region.putNodeEvent(new EvictedEventNode(fqn2, NodeEventType.ADD_NODE_EVENT));
      region.putNodeEvent(new EvictedEventNode(fqn3, NodeEventType.ADD_NODE_EVENT));

      algo_.process(region);
      assertEquals("Queue size #1: ", 3, algo_.getEvictionQueue().getNumberOfNodes());
      TestingUtil.sleepThread(2000);
      algo_.process(region);
      assertEquals("Queue size #2: ", 0, algo_.getEvictionQueue().getNumberOfNodes());
   }

   /**
    * MaxAgeSeconds = 2 with 3 nodes.
    *
    * @throws Exception
    */
   public void testMaxAgeSeconds2() throws Exception
   {
      Fqn fqn1 = Fqn.fromString("/a/b/c");
      Fqn fqn2 = Fqn.fromString("/a/b/d");
      Fqn fqn3 = Fqn.fromString("/a/b/e");
      Region region = regionManager_.getRegion("/a/b", true);
      LRUConfiguration config = (LRUConfiguration) region.getEvictionPolicyConfig();

      config.setMaxNodes(0);
      config.setTimeToLiveSeconds(0);
      config.setMaxAgeSeconds(2);
      region.putNodeEvent(new EvictedEventNode(fqn1, NodeEventType.ADD_NODE_EVENT));
      region.putNodeEvent(new EvictedEventNode(fqn2, NodeEventType.ADD_NODE_EVENT));
      region.putNodeEvent(new EvictedEventNode(fqn3, NodeEventType.ADD_NODE_EVENT));

      algo_.process(region);
      assertEquals("Queue size #1: ", 3, algo_.getEvictionQueue().getNumberOfNodes());
      TestingUtil.sleepThread(1000);
      algo_.process(region);
      assertEquals("Queue size #2: ", 3, algo_.getEvictionQueue().getNumberOfNodes());
      TestingUtil.sleepThread(1500);
      algo_.process(region);
      assertEquals("Queue size #3: ", 0, algo_.getEvictionQueue().getNumberOfNodes());
   }

   /**
    * Generic combo case.
    */
   public void testCombo1()
   {
      Fqn fqn1 = Fqn.fromString("/a/b/c");
      Fqn fqn2 = Fqn.fromString("/a/b/d");
      Fqn fqn3 = Fqn.fromString("/a/b/e");
      Fqn fqn4 = Fqn.fromString("/a/b/f");
      Region region = regionManager_.getRegion("/a/b", true);
      LRUConfiguration config = (LRUConfiguration) region.getEvictionPolicyConfig();

      config.setMaxNodes(2);
      config.setTimeToLiveSeconds(1);
      config.setMaxAgeSeconds(3);
      region.putNodeEvent(new EvictedEventNode(fqn1, NodeEventType.ADD_NODE_EVENT));
      region.putNodeEvent(new EvictedEventNode(fqn2, NodeEventType.ADD_NODE_EVENT));
      region.putNodeEvent(new EvictedEventNode(fqn4, NodeEventType.ADD_NODE_EVENT));

      try
      {
         algo_.process(region);
      }
      catch (EvictionException e)
      {
         fail("testMaxNode: process failed " + e);
         e.printStackTrace();
      }
      assertEquals("Queue size #1: ", 2, algo_.getEvictionQueue().getNumberOfNodes());
      region.putNodeEvent(new EvictedEventNode(fqn3, NodeEventType.ADD_NODE_EVENT));

      TestingUtil.sleepThread(2000);
      try
      {
         algo_.process(region);
      }
      catch (EvictionException e)
      {
         fail("testMaxNode: process failed " + e);
         e.printStackTrace();
      }
      assertEquals("Queue size #2: ", 1, algo_.getEvictionQueue().getNumberOfNodes());

      TestingUtil.sleepThread(3500);
      // visit the node now to prevent the idle time from doing the pruning - node still gets pruned but by
      // max age.
      region.putNodeEvent(new EvictedEventNode(fqn3, NodeEventType.VISIT_NODE_EVENT));
      try
      {
         algo_.process(region);
      }
      catch (EvictionException e)
      {
         fail("testMaxAge: process failed " + e);
         e.printStackTrace();
      }
      assertEquals("Queue size #3: ", 0, algo_.getEvictionQueue().getNumberOfNodes());
   }

   /**
    * Generic combo case with newly added node should be around.
    */
   public void testCombo2()
   {
      Fqn fqn1 = Fqn.fromString("/a/b/c");
      Fqn fqn2 = Fqn.fromString("/a/b/d");
      Fqn fqn3 = Fqn.fromString("/a/b/e");
      Region region = regionManager_.getRegion("/a/b", true);
      LRUConfiguration config = (LRUConfiguration) region.getEvictionPolicyConfig();

      config.setMaxNodes(2);
      config.setTimeToLiveSeconds(1);
      config.setMaxAgeSeconds(3);
      region.putNodeEvent(new EvictedEventNode(fqn1, NodeEventType.ADD_NODE_EVENT));
      region.putNodeEvent(new EvictedEventNode(fqn2, NodeEventType.ADD_NODE_EVENT));
      region.putNodeEvent(new EvictedEventNode(fqn2, NodeEventType.REMOVE_NODE_EVENT));
      try
      {
         algo_.process(region);
      }
      catch (EvictionException e)
      {
         fail("testMaxNode: process failed " + e);
         e.printStackTrace();
      }
      assertEquals("Queue size #1: ", 1, algo_.getEvictionQueue().getNumberOfNodes());
      region.putNodeEvent(new EvictedEventNode(fqn3, NodeEventType.ADD_NODE_EVENT));

      TestingUtil.sleepThread(2000);
      try
      {
         algo_.process(region);
      }
      catch (EvictionException e)
      {
         fail("testMaxNode: process failed " + e);
         e.printStackTrace();
      }
      assertEquals("Queue size #2: ", 1, algo_.getEvictionQueue().getNumberOfNodes());

      TestingUtil.sleepThread(3000);
      region.putNodeEvent(new EvictedEventNode(fqn3, NodeEventType.VISIT_NODE_EVENT));
      TestingUtil.sleepThread(500);
      try
      {
         algo_.process(region);
      }
      catch (EvictionException e)
      {
         fail("testCombo: process failed " + e);
         e.printStackTrace();
      }

      TestingUtil.sleepThread(1000); // Fairly random ?
      assertEquals("Queue size #3: ", 0, algo_.getEvictionQueue().getNumberOfNodes());
   }

   public void testEvictionSortOrder() throws Exception
   {
      Region region = regionManager_.getRegion("/a/b", true);
      LRUConfiguration config = (LRUConfiguration) region.getEvictionPolicyConfig();

      config.setMaxAgeSeconds(1000);
      config.setMaxNodes(0);
      config.setTimeToLiveSeconds(1000);

      for (int i = 0; i < 100; i++)
      {
         Fqn fqn = Fqn.fromString("/a/b/" + Integer.toString(i));
         region.putNodeEvent(new EvictedEventNode(fqn, NodeEventType.ADD_NODE_EVENT));
      }

      algo_.process(region);

      TestingUtil.sleepThread(5000);

      for (int i = 0; i < 100; i++)
      {
         Fqn fqn = Fqn.fromString("/a/b/" + Integer.toString(i));
         if (i % 2 == 0)
         {
            region.putNodeEvent(new EvictedEventNode(fqn, NodeEventType.VISIT_NODE_EVENT));
         }
      }

      algo_.process(region);

      LRUQueue queue = (LRUQueue) algo_.getEvictionQueue();

      NodeEntry ne;
      int count = 0;
      while ((ne = queue.getFirstLRUNodeEntry()) != null)
      {
         if (count < 50)
         {
            assertEquals(1, ne.getNumberOfNodeVisits());
         }
         else
         {
            assertEquals(2, ne.getNumberOfNodeVisits());
         }
         queue.removeNodeEntry(ne);
         count++;
      }

      for (int i = 0; i < 100; i++)
      {
         Fqn fqn = Fqn.fromString("/a/b/" + Integer.toString(i));
         region.putNodeEvent(new EvictedEventNode(fqn, NodeEventType.ADD_NODE_EVENT));
      }

      algo_.process(region);
      long lastCreateTimestamp = 0;
      while ((ne = queue.getFirstMaxAgeNodeEntry()) != null)
      {
         assertTrue(ne.getCreationTimeStamp() >= lastCreateTimestamp);
         lastCreateTimestamp = ne.getCreationTimeStamp();
         queue.removeNodeEntry(ne);
      }
   }

   void log(String msg)
   {
      System.out.println("-- " + msg);
   }

}
