/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.transaction;


import javax.transaction.NotSupportedException;
import javax.transaction.RollbackException;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;

/**
 * A dummy transaction manager that notifies registered listeners of the various phases of a 2PC so exceptions, etc. can be injected.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
public class NotifyingTransactionManager extends DummyTransactionManager implements TransactionManagerLookup
{

   private static final long serialVersionUID = -2994163352889758708L;

   private Notification notification;

   public void begin() throws SystemException, NotSupportedException
   {
      super.begin();
      try
      {
         System.out.println("Calling notification.notify()");
         notification.notify(getTransaction());
      }
      catch (RollbackException e)
      {
         e.printStackTrace();
      }
   }

   public TransactionManager getTransactionManager() throws Exception
   {
      return this;
   }

   public interface Notification
   {
      public void notify(Transaction tx) throws SystemException, RollbackException;
   }
   

   public Notification getNotification()
   {
      return notification;
   }
   

   public void setNotification(Notification notification)
   {
      this.notification = notification;
   }

}



